import copy
from typing import Any, AsyncIterator, Iterator

# Note: Import directly from langchain_core is not stable and generate some errors
from langchain_core.documents import Document

from .runnable_document_transformer import (
    _RunnableGeneratorDocumentTransformer,
)


class CopyDocumentTransformer(_RunnableGeneratorDocumentTransformer):
    """
    This class is a transparent transformation. The input is just sent back to
    the output.
    This is useful when you want to apply several transformations to a list of
    documents, and you also want to keep a copy of the original.
    With a DocumentTransformers, it's possible to do this in LCEL syntax.
    """

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        yield from (copy.deepcopy(doc) for doc in documents)

    async def _alazy_transform_documents(  # type: ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        async for doc in documents:
            yield copy.deepcopy(doc)



from .copy_transformer import CopyDocumentTransformer
from .document_transform_pipeline import DocumentTransformerPipeline
from .document_transformers import DocumentTransformers
from .generate_questions import GenerateQuestionsTransformer
from .runnable_document_transformer import LazyDocumentTransformer
from .sumarize_and_questions_transformer import (
    SummarizeAndQuestionsTransformer,
)
from .sumarize_transformer import SummarizeTransformer

__all__ = [
    "CopyDocumentTransformer",
    "DocumentTransformers",
    "DocumentTransformerPipeline",
    "GenerateQuestionsTransformer",
    "SummarizeAndQuestionsTransformer",
    "SummarizeTransformer",
    "LazyDocumentTransformer",
]



import copy
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    cast,
)

from langchain.chains import LLMChain
from langchain.output_parsers import PydanticOutputParser
from langchain_core.documents import Document
from langchain_core.language_models import BaseLanguageModel
from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import PromptTemplate
from langchain_core.pydantic_v1 import BaseModel

from langchain_rag.document_transformers.runnable_document_transformer import (
    _RunnableGeneratorDocumentTransformer,
)


def _default_get_input(doc: Document) -> Dict[str, Any]:
    """Return the context chain input."""
    return {
        "context": doc.page_content,
    }


class _SummarizeAndQuestions(BaseModel):
    summary: str
    """the document summary."""
    questions: List[str]
    """A list of questions"""


_default_parser: BaseOutputParser = PydanticOutputParser(
    pydantic_object=_SummarizeAndQuestions
)

_default_template = (
    "1. Given a text input, generate {nb_of_questions} questions from it in "
    "the same language. "
    "2. Summarize a text input in the same language.\n"
    "Context:\n"
    "```\n"
    "{context}\n"
    "```\n"
    "{format_instructions}\n"
)


def _get_default_chain_prompt() -> PromptTemplate:
    return PromptTemplate.from_template(
        template=_default_template,
        output_parser=_default_parser,
        partial_variables={
            "format_instructions": _default_parser.get_format_instructions()
        },
    )


class SummarizeAndQuestionsTransformer(_RunnableGeneratorDocumentTransformer):
    """Generate questions and summarize for each Documents."""

    llm_chain: LLMChain
    get_input: Callable[[Document], dict] = _default_get_input
    nb_of_questions: int = 3

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Compress page content of raw documents."""
        _callbacks = kwargs.get("callbacks", None)
        for doc in documents:
            _input = {
                **self.get_input(doc),
                **{"nb_of_questions": self.nb_of_questions},
            }
            output = cast(
                _SummarizeAndQuestions,
                self.llm_chain.predict(
                    callbacks=_callbacks,
                    **_input,
                ),
            )
            if not output:
                continue
            yield Document(page_content=output.summary, metadata=doc.metadata)
            for q in output.questions:
                metadata = copy.deepcopy(doc.metadata)
                metadata["transformer"] = self.__class__.__name__
                yield Document(page_content=q, metadata=metadata)

    async def _alazy_transform_documents(  # type:ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Compress page content of raw documents."""
        _callbacks = kwargs.get("callbacks", None)

        async for doc in documents:
            _input = {
                **self.get_input(doc),
                **{"nb_of_questions": self.nb_of_questions},
            }
            output = cast(
                _SummarizeAndQuestions,
                await self.llm_chain.apredict(
                    callbacks=_callbacks,
                    **_input,
                ),
            )
            if not output:
                continue
            yield Document(page_content=output.summary, metadata=doc.metadata)
            for q in output.questions:
                metadata = copy.deepcopy(doc.metadata)
                metadata["transformer"] = self.__class__.__name__
                yield Document(page_content=q, metadata=metadata)

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: Optional[PromptTemplate] = None,
        get_input: Optional[Callable[[Document], dict]] = None,
        nb_of_questions: int = 3,
        llm_chain_kwargs: Optional[dict] = None,
    ) -> "SummarizeAndQuestionsTransformer":
        """Initialize from LLM."""
        _prompt = prompt if prompt is not None else _get_default_chain_prompt()
        _get_input = get_input if get_input is not None else _default_get_input
        assert _prompt.output_parser
        llm_chain = LLMChain(
            llm=llm,
            prompt=_prompt,
            output_parser=cast(BaseOutputParser, _prompt.output_parser),
            **(llm_chain_kwargs or {}),
        )
        return cls(
            llm_chain=llm_chain, get_input=_get_input, nb_of_questions=nb_of_questions
        )



import copy
from typing import Any, AsyncIterator, Callable, Dict, Iterator, Optional

from langchain.chains import LLMChain
from langchain.output_parsers import NumberedListOutputParser
from langchain_core.documents import Document
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import PromptTemplate

from langchain_rag.document_transformers.runnable_document_transformer import (
    _RunnableGeneratorDocumentTransformer,
)


def _default_get_input(doc: Document) -> Dict[str, Any]:
    """Return the context chain input."""
    return {
        "context": doc.page_content,
    }


_default_template = """
Summarize a text input in the same language. 
Context:
```
{context}
```
"""
_default_format_instruction = NumberedListOutputParser()


def _get_default_chain_prompt() -> PromptTemplate:
    return PromptTemplate.from_template(
        template=_default_template,
    )


class SummarizeTransformer(_RunnableGeneratorDocumentTransformer):
    """Generate questions for each Documents."""

    llm_chain: LLMChain
    get_input: Callable[[Document], dict] = _default_get_input

    """LLM wrapper to use for compressing documents."""

    """Callable for constructing the chain input from the query and a Document."""

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        _callbacks = kwargs.get("callbacks", None)
        for doc in documents:
            _input = self.get_input(doc)
            output = self.llm_chain.predict(callbacks=_callbacks, **_input)
            if not output:
                continue
            metadata = copy.deepcopy(doc.metadata)
            metadata["transformer"] = self.__class__.__name__
            yield Document(
                page_content="SUMMARY:\n" + str(output).strip(), metadata=metadata
            )

    async def _alazy_transform_documents(  # type:ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        _callbacks = kwargs.get("callbacks", None)
        async for doc in documents:
            _input = self.get_input(doc)
            output = await self.llm_chain.apredict(callbacks=_callbacks, **_input)
            if not output:
                continue
            metadata = copy.deepcopy(doc.metadata)
            metadata["transformer"] = self.__class__.__name__
            yield Document(
                page_content="SUMMARY:\n" + str(output).strip(), metadata=metadata
            )

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: Optional[PromptTemplate] = None,
        get_input: Optional[Callable[[Document], dict]] = None,
        llm_chain_kwargs: Optional[dict] = None,
    ) -> "SummarizeTransformer":
        """Initialize from LLM."""
        _prompt = prompt if prompt is not None else _get_default_chain_prompt()
        _get_input = get_input if get_input is not None else _default_get_input
        llm_chain = LLMChain(
            llm=llm,
            prompt=_prompt,
            **(llm_chain_kwargs or {}),
        )
        return cls(llm_chain=llm_chain, get_input=_get_input)



import itertools
import sys
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Iterable,
    Iterator,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    no_type_check,
)

from langchain_core.documents import BaseDocumentTransformer, Document
from langchain_core.runnables.base import RunnableParallel, coerce_to_runnable

from langchain_rag.document_transformers.runnable_document_transformer import (
    _LEGACY,
    _RunnableGeneratorDocumentTransformer,
)

if sys.version_info.major > 3 or sys.version_info.minor >= 12:
    from itertools import batched  # type: ignore[attr-defined]
else:
    T = TypeVar("T")  # Only in python 3.12

    def batched(iterable: Iterable[T], n: int) -> Iterator[Tuple[T, ...]]:
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


BATCH_SIZE = 16


def _transform_documents_generator(
    documents: Iterator[Document],
    *,
    transformers: Sequence[_RunnableGeneratorDocumentTransformer],
) -> Iterator[Document]:
    Input = Union[AsyncIterator[Document], Iterator[Document]]
    steps = {
        f"transform_documents_{i}": transformer
        for i, transformer in enumerate(transformers)
    }
    # Implementation when all transformers are compatible with Runnable
    for batch in batched(documents, BATCH_SIZE):
        result = RunnableParallel[Input](steps=steps).invoke(iter(batch))
        for chunk in result["steps"].values():
            yield chunk


async def abatched(iterable: AsyncIterable[T], n: int) -> AsyncIterable[Tuple[T, ...]]:
    if n < 1:
        raise ValueError("n must be at least one")
    batch = []
    async for t in iterable:
        batch.append(t)
        if len(batch) > n:
            yield tuple(batch)
            batch.clear()
    if batch:
        yield tuple(batch)


class DocumentTransformers(_RunnableGeneratorDocumentTransformer):
    """
    Document transformer that uses a list of Transformers.
    Take each input document, and apply all transformations present in the
    `transformers` attribute.

    This is the basis for multiple transformations, using the plus operator.
    """

    class Config:
        arbitrary_types_allowed = True

    if _LEGACY:
        transformers: Sequence[BaseDocumentTransformer]
    else:
        transformers: Sequence[_RunnableGeneratorDocumentTransformer]  # type: ignore[no-redef]
    """List of document transformer that are applied in parallel."""

    def __add__(
        self,
        other: _RunnableGeneratorDocumentTransformer,
    ) -> "DocumentTransformers":
        """Compose this runnable with another object to create a RunnableSequence."""
        if isinstance(other, DocumentTransformers):
            return DocumentTransformers(
                transformers=list(other.transformers) + list(self.transformers),
            )
        else:
            if _LEGACY:
                return DocumentTransformers(
                    transformers=list(self.transformers) + [other]
                )
            else:
                return DocumentTransformers(
                    transformers=list(self.transformers)
                    + [cast(BaseDocumentTransformer, coerce_to_runnable(other))]
                )

    @no_type_check  # Bug in Mypy
    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an iterator of documents with the list of transformations.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        # Can be refactored to use parallelism
        if _LEGACY:
            # Implementation when all transformers are NOT compatible with Runnable
            # It's not compatible with lazy strategy. Load all documents and apply
            # all transformations.
            for batch in batched(documents, BATCH_SIZE):
                for t in self.transformers:
                    for doc in t.transform_documents(documents=list(batch)):
                        yield doc

        else:
            # Implementation with only the LCEL compatible transformers
            for batch in batched(documents, BATCH_SIZE):
                for t in self.transformers:
                    for doc in t.lazy_transform_documents(iter(batch)):
                        yield doc

    @no_type_check  # Bug in Mypy
    async def _alazy_transform_documents(
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents with a list
        of transformations.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An iterator of transformed Documents.
        """
        if _LEGACY:
            # Implementation when all transformers are NOT compatible with Runnable
            # It's not compatible with lazy strategy. Load all documents and apply
            # all transformations.
            async for batch in abatched(documents, BATCH_SIZE):
                for transformer in self.transformers:
                    for doc in await transformer.atransform_documents(batch):
                        yield doc
        else:
            # # Get a batch of documents, then apply each transformation by batch
            async for batch in abatched(documents, BATCH_SIZE):
                for transformer in self.transformers:
                    async for doc in transformer.alazy_transform_documents(iter(batch)):
                        yield doc



from itertools import chain
from typing import Any, AsyncIterator, Iterator, Sequence, cast

from langchain_core.documents import BaseDocumentTransformer, Document

from .runnable_document_transformer import (
    _RunnableGeneratorDocumentTransformer,
    to_async_iterator,
)


class DocumentTransformerPipeline(_RunnableGeneratorDocumentTransformer):
    class Config:
        arbitrary_types_allowed = True

    transformers: Sequence[BaseDocumentTransformer]
    """List of document transformers that are chained together and run in sequence."""
    # def __init__(self, transformers: Sequence[BaseDocumentTransformer]):
    #     self.transformers = transformers

    def _lazy_transform_documents_with_transformer(
        self,
        documents: Iterator[Document],
        transformer: BaseDocumentTransformer,
        **kwargs: Any,
    ) -> Iterator[Document]:
        lazy_results = []
        for _document in documents:
            if hasattr(transformer, "lazy_transform_documents"):
                lazy_results.append(
                    transformer.lazy_transform_documents([_document], **kwargs)
                )
            else:
                lazy_results.append(
                    iter(transformer.transform_documents([_document], **kwargs))
                )
        return chain(*lazy_results)

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        if not self.transformers:
            return iter(documents)
        for _transformer in self.transformers:
            documents = self._lazy_transform_documents_with_transformer(
                documents, transformer=_transformer
            )
        return documents

    async def _alazy_transform_documents_with_transformer(
        self,
        documents: AsyncIterator[Document],
        transformer: BaseDocumentTransformer,
        **kwargs: Any,
    ) -> AsyncIterator[Document]:
        lazy_results = []
        async for _document in documents:
            if hasattr(transformer, "lazy_transform_documents"):
                lazy_results.append(
                    await cast(
                        _RunnableGeneratorDocumentTransformer, transformer
                    ).alazy_transform_documents([_document], **kwargs)
                )
            else:
                lazy_results.append(
                    iter(await transformer.atransform_documents([_document], **kwargs))
                )
        return to_async_iterator(chain(*lazy_results))

    async def _alazy_transform_documents(  # type: ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        for _transformer in self.transformers:
            documents = await self._alazy_transform_documents_with_transformer(
                documents, transformer=_transformer
            )
        return documents



import contextlib
from pathlib import Path
from typing import (
    Any,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from langchain_core.stores import BaseStore
from sqlalchemy import (
    Column,
    Engine,
    PickleType,
    and_,
    create_engine,
)
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    AsyncSession,
    async_sessionmaker,
    create_async_engine,
)
from sqlalchemy.orm import (
    Mapped,
    Session,
    declarative_base,
    mapped_column,
    scoped_session,
    sessionmaker,
)

Base = declarative_base()


def items_equal(x: Any, y: Any) -> bool:
    return x == y


class Value(Base):  # type: ignore[valid-type,misc]
    """Table used to save values."""

    # ATTENTION:
    # Prior to modifying this table, please determine whether
    # we should create migrations for this table to make sure
    # users do not experience data loss.
    __tablename__ = "docstore"

    namespace: Mapped[str] = mapped_column(primary_key=True, index=True, nullable=False)
    key: Mapped[str] = mapped_column(primary_key=True, index=True, nullable=False)
    # value: Mapped[Any] = Column(type_=PickleType, index=False, nullable=False)
    value: Any = Column("earthquake", PickleType(comparator=items_equal))


# This is a fix of original SQLStore.
# This can will be removed when a PR will be merged.
class SQLStore(BaseStore[str, bytes]):
    """BaseStore interface that works on an SQL database.

    Examples:
        Create a SQLStore instance and perform operations on it:

        .. code-block:: python

            from langchain_rag.storage import SQLStore

            # Instantiate the SQLStore with the root path
            sql_store = SQLStore(namespace="test", db_url="sqllite://:memory:")

            # Set values for keys
            sql_store.mset([("key1", b"value1"), ("key2", b"value2")])

            # Get values for keys
            values = sql_store.mget(["key1", "key2"])  # Returns [b"value1", b"value2"]

            # Delete keys
            sql_store.mdelete(["key1"])

            # Iterate over keys
            for key in sql_store.yield_keys():
                print(key)

    """

    def __init__(
        self,
        *,
        namespace: str,
        db_url: Optional[Union[str, Path]] = None,
        engine: Optional[Union[Engine, AsyncEngine]] = None,
        engine_kwargs: Optional[Dict[str, Any]] = None,
        async_mode: bool = False,
    ):
        if db_url is None and engine is None:
            raise ValueError("Must specify either db_url or engine")

        if db_url is not None and engine is not None:
            raise ValueError("Must specify either db_url or engine, not both")

        _engine: Union[Engine, AsyncEngine]
        if db_url:
            if async_mode:
                _engine = create_async_engine(url=str(db_url), **(engine_kwargs or {}))
            else:
                _engine = create_engine(url=str(db_url), **(engine_kwargs or {}))
        elif engine:
            _engine = engine

        else:
            raise AssertionError("Something went wrong with configuration of engine.")

        _session_factory: Union[sessionmaker[Session], async_sessionmaker[AsyncSession]]
        if isinstance(_engine, AsyncEngine):
            _session_factory = async_sessionmaker(bind=_engine)
        else:
            _session_factory = sessionmaker(bind=_engine)

        self.engine = _engine
        self.dialect = _engine.dialect.name
        self.session_factory = _session_factory
        self.namespace = namespace

    def create_schema(self) -> None:
        Base.metadata.create_all(self.engine)

    def drop(self) -> None:
        Base.metadata.drop_all(bind=self.engine.connect())

    # async def amget(self, keys: Sequence[K]) -> List[Optional[V]]:
    #     result = {}
    #     async with self._make_session() as session:
    #         async with session.begin():
    #             for v in session.query(Value).filter(
    #                     and_(
    #                         Value.key.in_(keys),
    #                         Value.namespace == self.namespace,
    #                     )
    #             ):
    #                 result[v.key] = v.value
    #     return [result.get(key) for key in keys]

    def mget(self, keys: Sequence[str]) -> List[Optional[bytes]]:
        result = {}

        with self._make_session() as session:
            for v in session.query(Value).filter(  # type: ignore
                and_(
                    Value.key.in_(keys),
                    Value.namespace == self.namespace,
                )
            ):
                result[v.key] = v.value
        return [result.get(key) for key in keys]

    # async def amset(self, key_value_pairs: Sequence[Tuple[K, V]]) -> None:
    #     async with self._make_session() as session:
    #         async with session.begin():
    #             # await self._amdetete([key for key, _ in key_value_pairs], session)
    #             session.add_all([Value(namespace=self.namespace,
    #                                    key=k,
    #                                    value=v) for k, v in key_value_pairs])
    #             session.commit()

    def mset(self, key_value_pairs: Sequence[Tuple[str, bytes]]) -> None:
        # try:
        with self._make_session() as session:
            self._mdetete([key for key, _ in key_value_pairs], session)
            session.add_all(
                [
                    Value(namespace=self.namespace, key=k, value=v)
                    for k, v in key_value_pairs
                ]
            )

    def _mdetete(self, keys: Sequence[str], session: Session) -> None:
        with session.begin():
            session.query(Value).filter(  # type: ignore
                and_(
                    Value.key.in_(keys),
                    Value.namespace == self.namespace,
                )
            ).delete()

    # async def _amdetete(self, keys: Sequence[str], session: Session) -> None:
    #     await session.query(Value).filter(
    #         and_(
    #             Value.key.in_(keys),
    #             Value.namespace == self.namespace,
    #         )
    #     ).delete()

    def mdelete(self, keys: Sequence[str]) -> None:
        with self._make_session() as session:
            self._mdetete(keys, session)
            session.commit()

    # async def amdelete(self, keys: Sequence[str]) -> None:
    #     with self._make_session() as session:
    #         await self._mdelete(keys, session)
    #         session.commit()

    def yield_keys(self, *, prefix: Optional[str] = None) -> Iterator[str]:
        with self._make_session() as session:
            for v in session.query(Value).filter(  # type: ignore
                Value.namespace == self.namespace
            ):
                yield str(v.key)
            session.close()

    @contextlib.contextmanager
    def _make_session(self) -> Generator[Session, None, None]:
        """Create a session and close it after use."""

        if isinstance(self.session_factory, async_sessionmaker):
            raise AssertionError("This method is not supported for async engines.")

        session = scoped_session(self.session_factory)()
        try:
            yield session
        finally:
            session.commit()

    # @contextlib.asynccontextmanager
    # async def _amake_session(self) -> AsyncGenerator[AsyncSession, None]:
    #     """Create a session and close it after use."""
    #
    #     if not isinstance(self.session_factory, async_sessionmaker):
    #         raise AssertionError("This method is not supported for sync engines.")
    #
    #     async with self.session_factory() as session:
    #         yield session



from typing import Any, Iterable, List, Optional, Tuple, Type, TypeVar

# Note: Import directly from langchain_core is not stable and generate some errors
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore

VST = TypeVar("VST", bound="WrapperVectorStore")
T = TypeVar("T")


class WrapperVectorStore(VectorStore):
    """
    A generic wrapper for vectorstore.
    Can be known by SelfQueryRetriever to generate the target request.
    """

    def __init__(self, *, vectorstore: VectorStore):
        self.vectorstore = vectorstore

    @property
    def embeddings(self) -> Optional[Embeddings]:
        return self.vectorstore.embeddings

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> List[str]:
        return self.vectorstore.add_texts(texts=texts, metadatas=metadatas, **kwargs)

    async def aadd_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> List[str]:
        return await self.vectorstore.aadd_texts(
            texts=texts, metadatas=metadatas, **kwargs
        )

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        return self.delete(ids=ids, **kwargs)

    async def adelete(
        self, ids: Optional[List[str]] = None, **kwargs: Any
    ) -> Optional[bool]:
        return await self.vectorstore.adelete(ids=ids, **kwargs)

    def add_documents(self, documents: List[Document], **kwargs: Any) -> List[str]:
        return self.vectorstore.add_documents(documents=documents, **kwargs)

    async def aadd_documents(
        self, documents: List[Document], **kwargs: Any
    ) -> List[str]:
        return await self.vectorstore.aadd_documents(documents=documents, **kwargs)

    def search(self, query: str, search_type: str, **kwargs: Any) -> List[Document]:
        return self.vectorstore.search(query=query, search_type=search_type, **kwargs)

    async def asearch(
        self, query: str, search_type: str, **kwargs: Any
    ) -> List[Document]:
        return await self.vectorstore.asearch(
            query=query, search_type=search_type, **kwargs
        )

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return self.vectorstore.similarity_search(query=query, k=k, **kwargs)

    def similarity_search_with_score(
        self, *args: Any, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        return self.vectorstore.similarity_search_with_score(*args, **kwargs)

    def similarity_search_with_relevance_scores(
        self,
        query: str,
        k: int = 4,
        **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        return self.vectorstore.similarity_search_with_relevance_scores(
            query=query, k=k, **kwargs
        )

    async def asimilarity_search_with_relevance_scores(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        return await self.vectorstore.asimilarity_search_with_relevance_scores(
            query=query, k=k, **kwargs
        )

    async def asimilarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return await self.vectorstore.asimilarity_search(query=query, k=k, **kwargs)

    def similarity_search_by_vector(
        self, embedding: List[float], k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return self.vectorstore.similarity_search_by_vector(
            embedding=embedding, k=k, **kwargs
        )

    async def asimilarity_search_by_vector(
        self, embedding: List[float], k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return await self.vectorstore.asimilarity_search_by_vector(
            embedding=embedding, k=k, **kwargs
        )

    def max_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        return self.vectorstore.max_marginal_relevance_search(
            query=query, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )

    async def amax_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        return await self.vectorstore.amax_marginal_relevance_search(
            query=query, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )

    def max_marginal_relevance_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        return self.vectorstore.max_marginal_relevance_search_by_vector(
            embedding=embedding, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )

    async def amax_marginal_relevance_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        return await self.vectorstore.amax_marginal_relevance_search_by_vector(
            embedding=embedding, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )

    @classmethod
    def from_texts(
        cls: Type[VST],
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> VST:
        vectorstore_cls = kwargs.pop("vectorstore_cls")
        return cls(
            vectorstore=vectorstore_cls.from_texts(
                texts=texts, embedding=embedding, metadatas=metadatas, **kwargs
            )
        )


if __name__.startswith("langchain_"):

    def _hack() -> None:
        """
        This is a hack, to dynamically path the _get_builtin_translator()
        if we use the component before accepting the pull-request
        """
        from langchain.retrievers.self_query import base as to_patch

        old_get_builtin_translator = to_patch._get_builtin_translator

        def patch_get_builtin_translator(vectorstore: VectorStore) -> Any:
            if isinstance(vectorstore, WrapperVectorStore):
                return patch_get_builtin_translator(vectorstore.vectorstore)
            return old_get_builtin_translator(vectorstore)

        to_patch._get_builtin_translator = patch_get_builtin_translator

    _hack()



import asyncio
import threading
import time
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Iterable,
    Iterator,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    no_type_check,
)

from langchain_core.documents import BaseDocumentTransformer, Document
from langchain_core.pydantic_v1 import BaseModel
from langchain_core.runnables import Runnable, RunnableConfig
from langchain_core.runnables.base import coerce_to_runnable

if TYPE_CHECKING:
    from .document_transformers import DocumentTransformers
else:
    DocumentTransformers = Any

_LEGACY = True  # Use legacy langchain transformer interface

"""
    We propose an alternative way of making transformers compatible with LCEL.
    The first keeps the current protocol (RunnableDocumentTransformer).
    The second takes advantage of this to propose
    a lazy approach to transformations (_RunnableGeneratorDocumentTransformer).
    It's better for the memory, pipeline, etc.

    Now, it's possible to create a pipeline of transformer like:
    Example:
    ..code - block:: python
    class UpperTransformer(_RunnableGeneratorDocumentTransformer):
        def lazy_transform_documents(
                self,
                documents: Iterator[Document],
                **kwargs: Any
        ) -> Iterator[Document]:
            ...

        async def alazy_transform_documents(
                self,
                documents: Union[AsyncIterator[Document],Iterator[Document]],
                **kwargs: Any
        ) -> AsyncIterator[Document]:
            ...

    runnable = (UpperTransformer() | LowerTransformer())
    result = list(runnable.invoke(documents))
"""

T = TypeVar("T")


async def to_async_iterator(iterator: Iterator[T]) -> AsyncIterator[T]:
    """Convert an iterable to an async iterator."""
    for item in iterator:
        yield item


_DONE = ""
_TIMEOUT = 1


def to_sync_iterator(async_iterable: AsyncIterator[T], maxsize: int = 0) -> Iterator[T]:
    def _run_coroutine(
        loop: asyncio.AbstractEventLoop,
        async_iterable: AsyncIterator[T],
        queue: asyncio.Queue,
    ) -> None:
        async def _consume_async_iterable(
            async_iterable: AsyncIterator[T], queue: asyncio.Queue
        ) -> None:
            async for x in async_iterable:
                await queue.put(x)

            await queue.put(_DONE)

        loop.run_until_complete(_consume_async_iterable(async_iterable, queue))

    queue: asyncio.Queue[T] = asyncio.Queue(maxsize=maxsize)
    loop = asyncio.new_event_loop()

    t = threading.Thread(target=_run_coroutine, args=(loop, async_iterable, queue))
    t.daemon = True
    t.start()

    while True:
        if not queue.empty():
            x = queue.get_nowait()

            if x is _DONE:
                break
            else:
                yield x
        else:
            time.sleep(_TIMEOUT)

    t.join()


Input = Union[AsyncIterator[Document], Iterator[Document], Sequence[Document]]
Output = Union[AsyncIterator[Document], Iterator[Document]]


class LazyDocumentTransformer(BaseDocumentTransformer):
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        # Convert lazy to classical transformation
        return list(self.lazy_transform_documents(iter(documents), **kwargs))

    async def atransform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        return [
            doc
            async for doc in self.alazy_transform_documents(iter(documents), **kwargs)
        ]

    @abstractmethod
    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        """Transform an iterator of documents.

        Args:
            documents: A sequence of Documents to be transformed.

        Returns:
            An iterator oftransformed Documents.
        """
        raise NotImplementedError()

    @abstractmethod
    async def _alazy_transform_documents(  # type: ignore
        self, documents: AsyncIterator[Document], **kwargs: Any
    ) -> AsyncIterator[Document]:
        raise NotImplementedError()

    @no_type_check  # Bug in Mypy
    async def alazy_transform_documents(
        self,
        documents: Input,
        **kwargs: Any,
    ) -> AsyncIterator[Document]:
        """Asynchronously transform an iterator of documents.

        Args:
            documents: An iterator of Documents to be transformed.

        Returns:
            An iterator of transformed Documents.
        """
        if isinstance(documents, Iterable):
            async_documents = to_async_iterator(iter(documents))
        elif isinstance(documents, AsyncIterator):
            async_documents = documents
        else:
            async_documents = to_async_iterator(documents)

        async for doc in self._alazy_transform_documents(async_documents):
            yield doc


class _RunnableGeneratorDocumentTransformer(
    Runnable[
        Input,
        Output,
    ],
    LazyDocumentTransformer,
    BaseModel,  # Pydantic v2
    ABC,
):
    # This class has a new LCEL-compatible BaseDocumentTransformer proposal.
    # It is intended to replace it once all the impacts have been processed.
    """
    Runnable Document Transformer with lazy transformation.

    You can compose a list of transformations with the *or* operator.

        .. code-block:: python

            runnable=TokenTextSplitter(...) | CharacterTextSplitter(...)
            docs = list(runnable.invoke(input_docs))

    To apply multiple transformations with the same input, use the *plus* operator.

        .. code-block:: python

            runnable=TokenTextSplitter(...) + CharacterTextSplitter(...)
            docs = list(runnable.invoke(input_docs))

    and, you can combine these two operator


        .. code-block:: python

            runnable=(
                (TokenTextSplitter(...) | CharacterTextSplitter(...) ) +
                CharacterTextSplitter(...))
            docs = list(runnable.invoke(input_docs))

    > This class is a transition class for proposing lazy transformers,
    > compatible with LCEL.
    > Later, it can be integrated into BaseDocumentTransformer
    >if you agree to add a lazy approach to transformations.
    >All subclass of BaseDocumentTransformer must be updated to be compatible with this.
    """

    def __add__(
        self,
        other: "_RunnableGeneratorDocumentTransformer",
    ) -> "DocumentTransformers":
        """
        Compose this runnable with another object to create a
        RunnableTransformers, with several transformations applied to the
        same documents .
        """
        from .document_transformers import DocumentTransformers

        if isinstance(other, DocumentTransformers):
            return DocumentTransformers(
                transformers=list(other.transformers) + [self],
            )
        else:
            if _LEGACY:
                return DocumentTransformers(transformers=[self, other])
            else:
                return DocumentTransformers(
                    transformers=[
                        self,
                        cast(BaseDocumentTransformer, coerce_to_runnable(other)),
                    ]
                )

    if _LEGACY:

        def __radd__(
            self,
            other: "DocumentTransformers",
        ) -> "_RunnableGeneratorDocumentTransformer":
            return self.__add__(other)

    def invoke(
        self,
        input: Input,
        config: Optional[RunnableConfig] = None,
        **kwargs: Any,
    ) -> Output:
        if isinstance(input, AsyncIterator):
            raise ValueError("Use ainvoke() with async iterator")
        config = config or {}

        if hasattr(self, "lazy_transform_documents"):
            iterator = iter(input) if isinstance(input, Sequence) else input
            return self.lazy_transform_documents(iterator, **config)

        # Default implementation, without generator/iterator
        if isinstance(input, Sequence):
            return iter(self.transform_documents(input, **config))
        else:
            return iter(self.transform_documents(list(input), **config))

    async def ainvoke(
        self,
        input: Input,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Output:
        # # Default implementation, without generator
        config = config or {}

        if hasattr(self, "lazy_transform_documents"):
            iterator = iter(input) if isinstance(input, Sequence) else input
            return self.alazy_transform_documents(iterator, **config)

        # Default implementation, without generator/iterator
        if isinstance(input, Sequence):
            return iter(await self.atransform_documents(input, **config))
        elif isinstance(input, AsyncIterator):
            return iter(
                await self.atransform_documents([doc async for doc in input], **config)
            )
        else:
            return iter(await self.atransform_documents(list(input), **config))



import hashlib
import uuid
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
)

from langchain.storage import EncoderBackedStore
from langchain_core.documents import BaseDocumentTransformer, Document
from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import BaseModel, Extra, Field
from langchain_core.stores import BaseStore
from langchain_core.vectorstores import VectorStore, VectorStoreRetriever
from sqlalchemy import (
    Engine,
)
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
)

from .wrapper_vectorstore import WrapperVectorStore

# %%
VST = TypeVar("VST", bound="VectorStore")


def _get_source_id_assigner(
    source_id_key: Union[str, Callable[[Document], str], None],
) -> Callable[[Document], Union[str, None]]:
    """Get the source id from the document."""
    if source_id_key is None:
        return lambda doc: None
    elif isinstance(source_id_key, str):
        return lambda doc: doc.metadata[source_id_key]
    elif callable(source_id_key):
        return source_id_key
    else:
        raise ValueError(
            f"source_id_key should be either None, a string or a callable. "
            f"Got {source_id_key} of type {type(source_id_key)}."
        )


class RAGVectorStore(BaseModel, WrapperVectorStore):
    """Retrieve small chunks then retrieve their parent documents.

    When splitting documents for retrieval, there are often conflicting desires:

    1. You may want to have small documents, so that their embeddings can most
        accurately reflect their meaning. If too long, then the embeddings can
        lose meaning.
    2. You want to have long enough documents that the context of each chunk is
        retained.

    The ParentDocumentRetriever strikes that balance by splitting and storing
    small chunks of data. During retrieval, it first fetches the small chunks
    but then looks up the parent ids for those chunks and returns those larger
    documents.

    Note that "parent document" refers to the document that a small chunk
    originated from. This can either be the whole raw document OR a larger
    chunk.

    Examples:

        .. code-block:: python

            # Imports
            from langchain_community.vectorstores import Chroma
            from langchain_openai.embeddings import OpenAIEmbeddings
            from langchain_core.text_splitter import RecursiveCharacterTextSplitter
            from langchain.storage import InMemoryStore

            # This text splitter is used to create the parent documents
            parent_splitter = RecursiveCharacterTextSplitter(chunk_size=2000)
            # This text splitter is used to create the child documents
            # It should create documents smaller than the parent
            child_splitter = RecursiveCharacterTextSplitter(chunk_size=400)
            # The vectorstore to use to index the child chunks
            vectorstore = Chroma(embedding_function=OpenAIEmbeddings())
            # The storage layer for the parent documents
            store = InMemoryStore()

            # Initialize the retriever
            retriever = ParentDocumentRetriever(
                vectorstore=vectorstore,
                docstore=store,
                child_splitter=child_splitter,
                parent_splitter=parent_splitter,
            )
    """

    class Config:
        extra = Extra.forbid
        arbitrary_types_allowed = True

    vectorstore: VectorStore
    """The real vectorstore for saving chunks"""
    docstore: BaseStore[str, Union[Document, List[str]]]
    """The storage layer for the parent documents"""

    source_id_key: Union[str, Callable[[Document], str], None] = "source"
    """The metadata to identify the id of the parents """

    chunk_id_key: str = "_chunk_id"
    """The metadata to identify the chunk. Add an id if the chunk can not have one """

    child_ids_key: str = "_child_ids"
    """Contain a list with the vectorstore id for all 
    corresponding transformed chunk."""

    search_type: str = "similarity"
    """Type of search to perform. Defaults to "similarity"."""

    search_kwargs: dict = Field(default_factory=dict)
    """Keyword arguments to pass to the search function."""

    chunk_transformer: Optional[BaseDocumentTransformer] = None
    """The transformer to use to create child documents."""

    parent_transformer: Optional[BaseDocumentTransformer] = None
    """The transformer to use to create parent documents.
    If none, then the parent documents will be the raw documents passed in."""

    coef_trunk_k: int = 3
    """If search_kwargs["k"] is not set, use k * coef_trunk_k to loads trunks
    candidates.
    """

    def _get_trunk_from_sub_docs(
        self, sub_docs: List[Document], **kwargs: Any
    ) -> List[Document]:
        if self.chunk_transformer:
            ids = []
            for d in sub_docs:
                if d.metadata[self.chunk_id_key] not in ids:
                    ids.append(d.metadata[self.chunk_id_key])
            docs = cast(List[Document], self.docstore.mget(ids))
            result = [d for d in docs if d is not None]
        else:
            result = sub_docs
        if "k" in kwargs:
            return result[: kwargs["k"]]
        else:
            return result

    def _update_score_of_chunk(
        self, sub_chunks_and_score: List[Tuple[Document, float]]
    ) -> List[Tuple[Document, float]]:
        if not self.chunk_transformer:
            return sub_chunks_and_score
        ids = []
        scores: Dict[str, float] = {}
        key = self.chunk_id_key
        for d, s in sub_chunks_and_score:
            if d.metadata[key] not in ids:
                id = d.metadata[key]
                ids.append(id)
                chunk_s = scores.get(id, 1.0)
                scores[id] = min(chunk_s, s)
        docs = cast(Sequence[Document], self.docstore.mget(ids))
        map_ids = {doc.metadata[key]: i for doc, i in zip(docs, ids) if doc}
        return sorted(
            [(d, scores[map_ids[d.metadata[key]]]) for d in docs if d is not None],
            key=lambda x: x[1],
        )

    def _get_trunk_from_sub_docs_and_score(
        self, sub_docs_and_score: List[Tuple[Document, float]], k: int, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        if self.chunk_transformer:
            result = self._update_score_of_chunk(sub_docs_and_score)
        else:
            result = sub_docs_and_score
        if self.search_kwargs.get("k", k) < k:
            raise ValueError("The search_kwargs['k'] must be >= 'k'")
        return result[:k]

    def as_retriever(
        self, search_type: str = "similarity", search_kwargs: dict = {}, **kwargs: Any
    ) -> VectorStoreRetriever:
        if not self.chunk_transformer:
            return self.vectorstore.as_retriever(
                search_type=search_type, search_kwargs=search_kwargs, *kwargs
            )

        retriever = VectorStoreRetriever(
            vectorstore=self, search_type=search_type, search_kwargs=search_kwargs
        )
        return retriever

    def add_documents(
        self,
        documents: List[Document],
        *,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Adds documents to the docstore and vectorstores.

        Args:
            documents: List of documents to add
            ids: Optional list of ids for documents. If provided should be the same
                length as the list of documents. Can provided if parent documents
                are already in the document store and you don't want to re-add
                to the docstore. If not provided, random UUIDs will be used as
                ids.
        """
        source_id_key_get = _get_source_id_assigner(self.source_id_key)
        chunk_ids = None
        map_doc_ids: Dict[Any, str] = {}
        if self.parent_transformer:
            if ids:  # It's the parent ids
                if len(documents) != len(ids):
                    raise ValueError(
                        "Got uneven list of documents and ids. "
                        "If `ids` is provided, should be same length as `documents`."
                    )

                for id, doc in zip(ids, documents):
                    key = source_id_key_get(doc)
                    if key in map_doc_ids:
                        raise ValueError(
                            f"Got multiple documents with the same ids `{key}"
                        )
                    map_doc_ids[key] = id

            else:
                ids = []
                for doc in documents:
                    # Some docstore refuse some characters in the id.
                    # We convert the id to hash
                    doc_id = source_id_key_get(doc)
                    hash_id = hashlib.sha256(str(doc_id).encode("utf-8")).hexdigest()
                    ids.append(hash_id)
                    map_doc_ids[doc_id] = hash_id
            self.delete(ids=list(map_doc_ids.values()))
        else:
            chunk_ids = ids
            if chunk_ids:
                self.delete(ids=chunk_ids)
            ids = None

        if self.parent_transformer:
            if hasattr(self.parent_transformer, "lazy_transform_documents"):
                chunk_documents = list(
                    self.parent_transformer.lazy_transform_documents(iter(documents))
                )
            else:
                chunk_documents = list(
                    self.parent_transformer.transform_documents(documents)
                )
        else:
            chunk_documents = documents

        if chunk_ids is None:
            # Generate an id for each chunk, or use the ids
            # Put the associated chunk id after the transformation.
            # Then, it's possible to retrieve the original chunk with this
            # transformation.
            # for chunk in chunk_documents
            #     if self.chunk_id_key not in chunk.metadata:
            #         chunk.metadata[self.chunk_id_key]=str(uuid.uuid4())
            chunk_ids = [
                chunk.metadata.get(self.chunk_id_key, str(uuid.uuid4()))
                for chunk in chunk_documents
            ]

        chunk_ids_for_doc: Dict[str, List[str]] = {}
        if self.parent_transformer:
            # Associate each chunk with the parent
            for chunk_id, chunk_document in zip(chunk_ids, chunk_documents):
                doc_id = map_doc_ids[source_id_key_get(chunk_document)]
                list_of_chunk_ids = chunk_ids_for_doc.get(doc_id, [])
                list_of_chunk_ids.append(chunk_id)
                chunk_ids_for_doc[doc_id] = list_of_chunk_ids
                if self.chunk_id_key not in chunk_document.metadata:
                    chunk_document.metadata[self.chunk_id_key] = chunk_id

        full_chunk_docs = []
        if not self.chunk_transformer:
            self.vectorstore.add_documents(documents=chunk_documents, ids=chunk_ids)
        else:
            for chunk_id, chunk_doc in zip(chunk_ids, chunk_documents):
                all_transformed_chunk: Sequence[
                    Document
                ] = self.chunk_transformer.transform_documents(
                    [chunk_doc]
                )  # PPR: use multiple documents?
                # If in transformed chunk, add the id of the associated chunk
                for transformed_chunk in all_transformed_chunk:
                    transformed_chunk.metadata[self.chunk_id_key] = chunk_id
                # Save the transformed versions
                transformed_persistance_ids = self.vectorstore.add_documents(
                    list(all_transformed_chunk)
                )
                # Inject id of transformed ids in the chuck document
                chunk_doc.metadata[self.child_ids_key] = ",".join(
                    transformed_persistance_ids
                )
                # Prepare the mset in docstore
                full_chunk_docs.append((chunk_id, chunk_doc))

            # Add the chunks in docstore.
            # In the retriever, it's this instances to return
            # in metadata[child_ids_key], it's possible to find the id of all
            # transformed versions
            self.docstore.mset(full_chunk_docs)

        if self.parent_transformer:
            # With the *parent* mode, for each parent document,
            # we must save the id of all chunk.
            # Then, it's possible to remove/update all chunk when the parent document
            # was updated.
            # Save the parent association with all chunk
            ids = cast(List[str], ids)
            mset_values: List[Tuple[str, List[str]]] = []
            for parent_id, doc in zip(ids, documents):
                mset_values.append((parent_id, chunk_ids_for_doc[parent_id]))
            self.docstore.mset(mset_values)
            return ids
        else:
            return chunk_ids

    async def aadd_documents(
        self,
        documents: List[Document],
        *,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Adds documents to the docstore and vectorstores.

        Args:
            documents: List of documents to add
            ids: Optional list of ids for documents. If provided should be the same
                length as the list of documents. Can provided if parent documents
                are already in the document store and you don't want to re-add
                to the docstore. If not provided, random UUIDs will be used as
                ids.
        """
        source_id_key_get = _get_source_id_assigner(self.source_id_key)
        chunk_ids = None
        map_doc_ids: Dict[Any, str] = {}
        if self.parent_transformer:
            if ids:  # It's the parent ids
                if len(documents) != len(ids):
                    raise ValueError(
                        "Got uneven list of documents and ids. "
                        "If `ids` is provided, should be same length as `documents`."
                    )

                for id, doc in zip(ids, documents):
                    map_doc_ids[source_id_key_get(doc)] = id

            else:
                for doc in documents:
                    ids = []
                    for doc in documents:
                        # Some docstore refuse some characters in the id.
                        # We convert the id to hash
                        doc_id = source_id_key_get(doc)
                        hash_id = hashlib.sha256(
                            str(doc_id).encode("utf-8")
                        ).hexdigest()
                        ids.append(hash_id)
                        map_doc_ids[doc_id] = hash_id
            await self.adelete(ids=list(map_doc_ids.values()))
        else:
            chunk_ids = ids
            if chunk_ids:
                await self.adelete(ids=chunk_ids)
            ids = None

        if self.parent_transformer:
            if hasattr(self.parent_transformer, "alazy_transform_documents"):
                chunk_documents = list(
                    await self.parent_transformer.alazy_transform_documents(
                        iter(documents)
                    )
                )
            else:
                chunk_documents = list(
                    await self.parent_transformer.atransform_documents(documents)
                )
        else:
            chunk_documents = documents

        if chunk_ids is None:
            # Generate an id for each chunk, or use the ids
            # Put the associated chunk id after the transformation.
            # Then, it's possible to retrieve the original chunk with this
            # transformation.
            # for chunk in chunk_documents
            #     if self.chunk_id_key not in chunk.metadata:
            #         chunk.metadata[self.chunk_id_key]=str(uuid.uuid4())
            chunk_ids = [
                chunk.metadata.get(self.chunk_id_key, str(uuid.uuid4()))
                for chunk in chunk_documents
            ]

        chunk_ids_for_doc: Dict[str, List[str]] = {}
        if self.parent_transformer:
            # Associate each chunk with the parent
            for chunk_id, chunk_document in zip(chunk_ids, chunk_documents):
                doc_id = map_doc_ids[source_id_key_get(chunk_document)]
                list_of_chunk_ids = chunk_ids_for_doc.get(doc_id, [])
                list_of_chunk_ids.append(chunk_id)
                chunk_ids_for_doc[doc_id] = list_of_chunk_ids
                if self.chunk_id_key not in chunk_document.metadata:
                    chunk_document.metadata[self.chunk_id_key] = chunk_id

        full_chunk_docs = []
        if not self.chunk_transformer:
            await self.vectorstore.aadd_documents(
                documents=chunk_documents, ids=chunk_ids
            )
        else:
            for chunk_id, chunk_doc in zip(chunk_ids, chunk_documents):
                all_transformed_chunk: Sequence[
                    Document
                ] = await self.chunk_transformer.atransform_documents(
                    [chunk_doc]
                )  # PPR: use multiple documents?
                # If in transformed chunk, add the id of the associated chunk
                for transformed_chunk in all_transformed_chunk:
                    transformed_chunk.metadata[self.chunk_id_key] = chunk_id
                # Save the transformed versions
                transformed_persistance_ids = await self.vectorstore.aadd_documents(
                    list(all_transformed_chunk)
                )
                # Inject id of transformed ids in the chuck document
                chunk_doc.metadata[self.child_ids_key] = ",".join(
                    transformed_persistance_ids
                )
                # Prepare the mset in docstore
                full_chunk_docs.append((chunk_id, chunk_doc))

            # Add the chunks in docstore.
            # In the retriever, it's this instances to return
            # in metadata[child_ids_key], it's possible to find the id of all
            # transformed versions
            self.docstore.mset(full_chunk_docs)

        if self.parent_transformer:
            # With the *parent* mode, for each parent document,
            # we must save the id of all chunk.
            # Then, it's possible to remove/update all chunk when the parent document
            # was updated.
            # Save the parent association with all chunk
            ids = cast(List[str], ids)
            mset_values: List[Tuple[str, List[str]]] = []
            for parent_id, doc in zip(ids, documents):
                mset_values.append((parent_id, chunk_ids_for_doc[parent_id]))
            self.docstore.mset(mset_values)
            return ids
        else:
            return chunk_ids

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> Optional[bool]:
        if not ids:
            raise ValueError("ids must be set")
        if self.parent_transformer:
            if not ids:
                raise ValueError("ids must be set")
            lists_of_chunk_by_doc_ids = cast(List[List[str]], self.docstore.mget(ids))
            chunk_by_doc_ids: List[str] = []
            for list_of_ids in lists_of_chunk_by_doc_ids:
                if list_of_ids:
                    chunk_by_doc_ids.extend([id for id in list_of_ids])
        else:
            chunk_by_doc_ids = ids

        if not any(chunk_by_doc_ids):
            return False

        transformed_ids = set()
        if self.chunk_transformer:
            chunk_docs = cast(List[Document], self.docstore.mget(chunk_by_doc_ids))
            self.docstore.mdelete(chunk_by_doc_ids)
            for chunk_doc in chunk_docs:
                if chunk_doc:
                    transformed_ids.update(
                        chunk_doc.metadata[self.child_ids_key].split(",")
                    )
        if transformed_ids:
            self.vectorstore.delete(ids=list(transformed_ids))
        elif self.parent_transformer:
            return self.vectorstore.delete(ids=chunk_by_doc_ids)
        elif not self.parent_transformer and self.chunk_transformer:
            return len(transformed_ids) != 0
        else:
            return self.vectorstore.delete(ids=ids)
        return False

    async def adelete(
        self, ids: Optional[List[str]] = None, **kwargs: Any
    ) -> Optional[bool]:
        if not ids:
            raise ValueError("ids must be set")
        if self.parent_transformer:
            if not ids:
                raise ValueError("ids must be set")
            lists_of_chunk_by_doc_ids = cast(List[List[str]], self.docstore.mget(ids))
            chunk_by_doc_ids: List[str] = []
            for list_of_ids in lists_of_chunk_by_doc_ids:
                if list_of_ids:
                    chunk_by_doc_ids.extend([id for id in list_of_ids])
        else:
            chunk_by_doc_ids = ids

        if not any(chunk_by_doc_ids):
            return False

        transformed_ids = set()
        if self.chunk_transformer:
            chunk_docs = cast(List[Document], self.docstore.mget(chunk_by_doc_ids))
            self.docstore.mdelete(chunk_by_doc_ids)
            for chunk_doc in chunk_docs:
                if chunk_doc:
                    transformed_ids.update(
                        chunk_doc.metadata[self.child_ids_key].split(",")
                    )
        if transformed_ids:
            await self.vectorstore.adelete(ids=list(transformed_ids))
        elif self.parent_transformer:
            return await self.vectorstore.adelete(ids=chunk_by_doc_ids)
        elif not self.parent_transformer and self.chunk_transformer:
            return len(transformed_ids) != 0
        else:
            return await self.vectorstore.adelete(ids=ids)
        return False

    @classmethod
    def from_texts(
        cls: Type[VST],
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> VST:
        raise NotImplementedError("from_texts not implemented")

    # %% searches
    def _trunk_k(
        self, result: List[Document], kwargs: Dict[str, Any]
    ) -> List[Document]:
        if "k" in kwargs:
            return result[: kwargs["k"]]
        else:
            return result

    def search(self, query: str, search_type: str, **kwargs: Any) -> List[Document]:
        _search_kwargs = {**kwargs, **self.search_kwargs}
        subdocs = self.vectorstore.search(
            query=query, search_type=search_type, **_search_kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, **kwargs)

    async def asearch(
        self, query: str, search_type: str, **kwargs: Any
    ) -> List[Document]:
        _search_kwargs = {**kwargs, **self.search_kwargs}
        subdocs = await self.vectorstore.asearch(
            query=query, search_type=search_type, **_search_kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, **kwargs)

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return self.search(query=query, search_type="similarity", k=k, **kwargs)

    async def asimilarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        return await self.asearch(query=query, search_type="similarity", k=k, **kwargs)

    def similarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        _search_kwargs = {**kwargs, **self.search_kwargs}

        trunk_k: Optional[int] = self.search_kwargs.get("k")
        if not trunk_k:
            _search_kwargs["k"] = k * self.coef_trunk_k

        subdocs_and_score = self.vectorstore.similarity_search_with_score(
            query=query, **_search_kwargs
        )
        return self._get_trunk_from_sub_docs_and_score(subdocs_and_score, k=k)

    async def asimilarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        _search_kwargs = {**kwargs, **self.search_kwargs}

        trunk_k: Optional[int] = self.search_kwargs.get("k")
        if not trunk_k:
            _search_kwargs["k"] = k * self.coef_trunk_k

        subdocs_and_score = await self.vectorstore.asimilarity_search_with_score(
            query=query, **_search_kwargs
        )
        return self._get_trunk_from_sub_docs_and_score(subdocs_and_score, k=k)

    def similarity_search_with_relevance_scores(
        self,
        query: str,
        k: int = 4,
        **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        _search_kwargs = {**kwargs, **self.search_kwargs}
        subdocs_and_score = self.vectorstore.similarity_search_with_relevance_scores(
            query=query, **_search_kwargs
        )
        return self._update_score_of_chunk(subdocs_and_score)[:k]

    async def asimilarity_search_with_relevance_scores(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        _search_kwargs = {**kwargs, **self.search_kwargs}
        subdocs_and_score = (
            await self.vectorstore.asimilarity_search_with_relevance_scores(
                query=query, **_search_kwargs
            )
        )
        return self._update_score_of_chunk(subdocs_and_score)[:k]

    def similarity_search_by_vector(
        self, embedding: List[float], k: int = 4, **kwargs: Any
    ) -> List[Document]:
        subdocs = self.vectorstore.similarity_search_by_vector(
            embedding=embedding, k=k, **kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, k=k)

    async def asimilarity_search_by_vector(
        self, embedding: List[float], k: int = 4, **kwargs: Any
    ) -> List[Document]:
        subdocs = await self.vectorstore.asimilarity_search_by_vector(
            embedding=embedding, k=k, **kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, k=k)

    def max_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        subdocs = self.vectorstore.max_marginal_relevance_search(
            query=query, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, k=k)

    async def amax_marginal_relevance_search(
        self,
        query: str,
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        subdocs = await self.vectorstore.amax_marginal_relevance_search(
            query=query, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )
        return self._get_trunk_from_sub_docs(
            subdocs,
            k=k,
        )

    def max_marginal_relevance_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        subdocs = self.vectorstore.max_marginal_relevance_search_by_vector(
            embedding=embedding, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs, k=k)

    async def amax_marginal_relevance_search_by_vector(
        self,
        embedding: List[float],
        k: int = 4,
        fetch_k: int = 20,
        lambda_mult: float = 0.5,
        **kwargs: Any,
    ) -> List[Document]:
        subdocs = await self.vectorstore.amax_marginal_relevance_search_by_vector(
            embedding=embedding, k=k, fetch_k=fetch_k, lambda_mult=lambda_mult, **kwargs
        )
        return self._get_trunk_from_sub_docs(subdocs)

    @staticmethod
    def from_vs_in_memory(
        vectorstore: VectorStore,
        *,
        chunk_transformer: Optional[BaseDocumentTransformer] = None,
        parent_transformer: Optional[BaseDocumentTransformer] = None,
        source_id_key: Union[str, Callable[[Document], str]] = "source",
        **kwargs: Any,
    ) -> Tuple["RAGVectorStore", Dict[str, Any]]:
        from langchain.storage import InMemoryStore

        from ..indexes.memory_recordmanager import MemoryRecordManager

        record_manager = MemoryRecordManager(namespace="in-memory")
        docstore = InMemoryStore()
        vectorstore = RAGVectorStore(
            vectorstore=vectorstore,
            docstore=docstore,
            parent_transformer=parent_transformer,
            chunk_transformer=chunk_transformer,
            **kwargs,
        )
        return (
            vectorstore,
            {
                "record_manager": record_manager,
                "vector_store": vectorstore,
                "source_id_key": source_id_key,
            },
        )

    @staticmethod
    def from_vs_in_sql(
        vectorstore: VectorStore,
        engine: Optional[Union[Engine, AsyncEngine]] = None,
        engine_kwargs: Optional[Dict[str, Any]] = None,
        db_url: Optional[str] = None,
        namespace: str = "rag_vectorstore",
        *,
        chunk_transformer: Optional[BaseDocumentTransformer] = None,
        parent_transformer: Optional[BaseDocumentTransformer] = None,
        **kwargs: Any,
    ) -> Tuple["RAGVectorStore", Dict[str, Any]]:
        from langchain.indexes import SQLRecordManager

        docstore: BaseStore[str, Union[Document, List[str]]]
        if not db_url and not engine:
            raise ValueError("Set db_url or engine")
        if db_url:
            record_manager = SQLRecordManager(namespace=namespace, db_url=db_url)
            record_manager.create_schema()
            # This implementation is not correct now.
            # from langchain_community.storage.sql import SQLBaseStore
            # docstore = SQLBaseStore[Union[Document, List[str]]](
            #     collection_name=namespace,
            #     connection_string=db_url,
            # )
            import pickle

            from ..storage.sql_docstore import SQLStore

            sql_docstore = SQLStore(
                namespace=namespace,
                db_url=db_url,
            )
            sql_docstore.create_schema()
            docstore = EncoderBackedStore[str, Union[Document, List[str]]](
                store=sql_docstore,
                key_encoder=lambda x: x,
                value_serializer=pickle.dumps,
                value_deserializer=pickle.loads,
            )
        else:
            # Note: the PR https://github.com/langchain-ai/langchain/pull/15909
            # is not compatible with "engine"
            import pickle

            from ..storage.sql_docstore import SQLStore

            record_manager = SQLRecordManager(
                namespace=namespace, engine=engine, engine_kwargs=engine_kwargs
            )
            record_manager.create_schema()
            sql_docstore = SQLStore(
                namespace=namespace, engine=engine, engine_kwargs=engine_kwargs
            )
            sql_docstore.create_schema()
            docstore = EncoderBackedStore[str, Union[Document, List[str]]](
                store=sql_docstore,
                key_encoder=lambda x: x,
                value_serializer=pickle.dumps,
                value_deserializer=pickle.loads,
            )

        vectorstore = RAGVectorStore(
            vectorstore=vectorstore,
            docstore=docstore,
            parent_transformer=parent_transformer,
            chunk_transformer=chunk_transformer,
            **kwargs,
        )
        return (
            vectorstore,
            {
                "record_manager": record_manager,
                "vector_store": vectorstore,
                "source_id_key": kwargs.get("source_id_key", "source"),
            },
        )



from datetime import datetime
from typing import Any, Dict, List, Optional, Sequence

from langchain_community.indexes.base import RecordManager


class MemoryRecordManager(RecordManager):
    data: List[Dict[str, Any]] = []

    def __init__(self, namespace: str):
        super().__init__(namespace=namespace)

    def create_schema(self) -> None:
        pass

    async def acreate_schema(self) -> None:
        pass

    def get_time(self) -> float:
        return datetime.now().timestamp()

    def update(
        self,
        keys: Sequence[str],
        *,
        group_ids: Optional[Sequence[Optional[str]]] = None,
        time_at_least: Optional[float] = None,
    ) -> None:
        if group_ids is None:
            group_ids = [None] * len(keys)
        if len(keys) != len(group_ids):
            raise ValueError(
                f"Number of keys ({len(keys)}) does not match number of "
                f"group_ids ({len(group_ids)})"
            )

        update_time = self.get_time()
        if time_at_least and update_time < time_at_least:
            # Safeguard against time sync issues
            raise AssertionError(f"Time sync issue: {update_time} < {time_at_least}")

        records_to_upsert = [
            {
                "key": key,
                "namespace": self.namespace,
                "updated_at": update_time,
                "group_id": group_id,
            }
            for key, group_id in zip(keys, group_ids)
        ]
        self.delete_keys(keys)
        self.data.extend(records_to_upsert)

    def exists(self, keys: Sequence[str]) -> List[bool]:
        return [
            len(list(filter(lambda record: record["key"] == key, self.data))) == 1
            for key in keys
        ]

    def list_keys(
        self,
        *,
        before: Optional[float] = None,
        after: Optional[float] = None,
        group_ids: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
    ) -> List[str]:
        keys = [
            record["key"]
            for record in filter(
                lambda record: record["namespace"] == self.namespace
                and (not group_ids or record["group_id"] in group_ids)
                and (not before or record["updated_at"] < before)
                and (not after or record["updated_at"] > after),
                self.data,
            )
        ]
        return keys[:limit]

    def delete_keys(self, keys: Sequence[str]) -> None:
        self.data = list(
            filter(
                lambda record: record["namespace"] != self.namespace
                or record["key"] not in keys,
                self.data,
            )
        )

    # %% Async versions
    async def aget_time(self) -> float:
        return datetime.now().timestamp()

    async def aupdate(
        self,
        keys: Sequence[str],
        *,
        group_ids: Optional[Sequence[Optional[str]]] = None,
        time_at_least: Optional[float] = None,
    ) -> None:
        return self.update(keys=keys, group_ids=group_ids, time_at_least=time_at_least)

    async def aexists(self, keys: Sequence[str]) -> List[bool]:
        return self.exists(keys=keys)

    async def alist_keys(
        self,
        *,
        before: Optional[float] = None,
        after: Optional[float] = None,
        group_ids: Optional[Sequence[str]] = None,
        limit: Optional[int] = None,
    ) -> List[str]:
        return self.list_keys(
            before=before, after=after, group_ids=group_ids, limit=limit
        )

    async def adelete_keys(self, keys: Sequence[str]) -> None:
        return self.delete_keys(keys=keys)



from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    Optional,
    Sequence,
    cast,
)

from langchain.chains import LLMChain
from langchain_core.documents import Document
from langchain_core.language_models import BaseLanguageModel
from langchain_core.output_parsers import BaseOutputParser, NumberedListOutputParser
from langchain_core.prompts import PromptTemplate

from langchain_rag.document_transformers.runnable_document_transformer import (
    _RunnableGeneratorDocumentTransformer,
)


def _default_get_input(doc: Document) -> Dict[str, Any]:
    """Return the context chain input."""
    return {
        "context": doc.page_content,
    }


_default_template = """
Given a text input, generate {nb_of_questions} questions from it in the same language. 
Context:
```
{context}
```
{format_instructions}"""

_default_parser = NumberedListOutputParser()


def _get_default_chain_prompt() -> PromptTemplate:
    return PromptTemplate.from_template(
        template=_default_template,
        output_parser=_default_parser,
        partial_variables={
            "format_instructions": _default_parser.get_format_instructions()
        },
    )


class GenerateQuestionsTransformer(_RunnableGeneratorDocumentTransformer):
    """Generate questions for each Documents."""

    llm_chain: LLMChain
    get_input: Callable[[Document], dict] = _default_get_input
    nb_of_questions: int = 3

    """Callable for constructing the chain input from the query and a Document."""

    def lazy_transform_documents(
        self, documents: Iterator[Document], **kwargs: Any
    ) -> Iterator[Document]:
        _callbacks = kwargs.get("callbacks", None)
        max_retry = 3
        for doc in documents:  # PPR: work with batch?
            _input = {
                **self.get_input(doc),
                **{"nb_of_questions": self.nb_of_questions},
            }
            output = cast(
                Sequence[str], self.llm_chain.predict(callbacks=_callbacks, **_input)
            )
            if not output:
                if max_retry := max_retry - 1:
                    continue
                else:
                    return
            for question in output:
                yield Document(page_content=question, metadata=doc.metadata)

    async def _alazy_transform_documents(  # type: ignore
        self,
        documents: AsyncIterator[Document],
        **kwargs: Any,
    ) -> AsyncIterator[Document]:
        """Compress page content of raw documents asynchronously."""
        _callbacks = kwargs.get("callbacks", None)
        async for doc in documents:  # PPR: work with batch?
            _input = {
                **self.get_input(doc),
                **{"nb_of_questions": self.nb_of_questions},
            }
            output = cast(
                Sequence[str],
                await self.llm_chain.apredict(callbacks=_callbacks, **_input),
            )
            if not output:
                continue
            for question in output:
                yield Document(page_content=question, metadata=doc.metadata)

    @classmethod
    def from_llm(
        cls,
        llm: BaseLanguageModel,
        prompt: Optional[PromptTemplate] = None,
        get_input: Optional[Callable[[Document], dict]] = None,
        nb_of_questions: int = 3,
        llm_chain_kwargs: Optional[dict] = None,
    ) -> "GenerateQuestionsTransformer":
        """Initialize from LLM."""
        _prompt = prompt if prompt is not None else _get_default_chain_prompt()
        _get_input = get_input if get_input is not None else _default_get_input
        assert _prompt.output_parser
        llm_chain = LLMChain(
            llm=llm,
            prompt=_prompt,
            output_parser=cast(BaseOutputParser, _prompt.output_parser),
            **(llm_chain_kwargs or {}),
        )
        return cls(
            llm_chain=llm_chain, get_input=_get_input, nb_of_questions=nb_of_questions
        )
